from abc import ABCMeta, abstractmethod
from typing import Dict, List, Optional, Union, Any

import napalm.base.models as napalm_models
from . import models as umnet_models


class AbstractUMnetNapalm(metaclass=ABCMeta):
    """
    Abstract definition of umnet_napalm class.
    Some of these methods are expected to be inherited from
    NAPALM. Extensions to NAPALM are defined in vendor-specific child classes

    """

    @abstractmethod
    def __init__(
        self,
        hostname: str,
        username: str,
        password: str,
        timeout: int = 60,
        optional_args: Optional[Dict] = None,
    ) -> None:
        pass

    @abstractmethod
    def cli(
        self, commands: List[str], encoding: str = "text"
    ) -> Dict[str, Union[str, Dict[str, Any]]]:
        """Run a list of CLI command and get the results as a dict keyed
        on the command that you ran"""

    @abstractmethod
    def get_facts(self) -> napalm_models.FactsDict:
        """get basic device facts"""

    @abstractmethod
    def get_config(self) -> napalm_models.ConfigDict:
        """get configs"""

    @abstractmethod
    def get_lldp_neighbors(self) -> Dict[str, List[napalm_models.LLDPNeighborDict]]:
        """get lldp neighbors"""

    @abstractmethod
    def get_ip_interfaces(self) -> List[umnet_models.IPInterfaceDict]:
        """get IP interface information"""

    @abstractmethod
    def get_arp_table(self) -> List[napalm_models.ARPTableDict]:
        """get IP interface information"""

    @abstractmethod
    def get_active_routes(self) -> List[umnet_models.RouteDict]:
        """get active routes"""

    @abstractmethod
    def get_mpls_switching(self) -> List[umnet_models.MPLSDict]:
        """get mpls switching (the mpls forwarding table)"""

    @abstractmethod
    def get_vni_information(self) -> List[umnet_models.VNIDict]:
        """get vni to vlan and VRF mapping"""

    @abstractmethod
    def get_inventory(self) -> List[umnet_models.InventoryDict]:
        """get inventory items"""

    @abstractmethod
    def get_lag_interfaces(self) -> Dict[str, umnet_models.LagInterfaceDict]:
        """get lag interfaces"""



from io import StringIO
import re
from datetime import datetime
from ipaddress import ip_network, ip_interface, ip_address


class MutableTuple(list):
    """
    Class used to identify a tuple that's been parsed
    from a string (dynamically created)
    """


# valid top-level key characters for 'show system state'
VALID_KEY_CHARS = r"[\w\.\-_]"


def system_state_value_to_type(input_string: str) -> any:
    """
    Converts a string to an int, float, bool, date, or IP
    """

    # "None", "True", or "False"
    for match, out in [("None", None), ("True", True), ("False", False)]:
        if input_string == match:
            return out

    # Integer conversion
    if re.match(r"^\d+$", input_string):
        return int(input_string)

    # Decimal/float
    if re.match(r"\d+\.\d+$", input_string):
        return float(input_string)

    # datetime, format appears to be consistent
    try:
        return datetime.strptime(input_string, "%Y/%m/%d %H:%M:%S")
    except ValueError:
        pass

    # IP
    for ip_type in [ip_network, ip_interface, ip_address]:
        try:
            return ip_type(input_string)
        except ValueError:
            pass

    return input_string


def parse_system_state(system_state: str) -> dict:
    """
    Parses "show system state", either filtered or unfiltered,
    from a raw string into structured data.

    Note that "show system state" output consists of a long list of top-level
    entries that we're saving as a dictionary. Each entry can be a string (single or multi-line),
    or a complex nested object of dicts, lists, and tuples.
    """

    # strip off xml tags
    system_state = system_state.replace('<response status="success"><result>', "")
    system_state = system_state.replace("</result></response>", "")

    # top of our structure starts off as an empty dict
    curr_node = {}

    buffer = ""
    curr_key = ""

    # for keeping track of nested object values
    obj_stack = []

    # to flag whether we're reading in the name of a top key or
    # a nested key
    reading_key = False
    reading_top_key = True

    # handling multiline values require us to keep track of the last top-level
    # node we just saved
    prev_node = {}
    prev_key = ""
    multiline_value = False

    # stream the output into our logic one character at a time
    stream = StringIO(system_state)
    while True:

        # peel off next character
        c = stream.read(1)
        if not c:
            break

        # colon denotes the end of a top-level key (unless this is a multiline value)
        if c == ":" and reading_top_key and not multiline_value:
            curr_key = buffer
            buffer = ""
            reading_top_key = False

            next_c = stream.read(1)
            if next_c != " ":
                raise ValueError(
                    f"Missing a space after reading top level key {curr_key} at idx {stream.tell()}"
                )
            continue

        # If we're at the top level and we're reading in the key but we hit an invalid character for keys,
        # we need to assume this is a multi-line value for the last top-level entry
        if not (re.match(VALID_KEY_CHARS, c)) and reading_top_key:
            multiline_value = True
            buffer += c
            continue

        # A newline almost always means we're done reading in a top-level entry's value...
        if c == "\n" and len(obj_stack) == 0:

            # unless we're reading in a multiline value, in which case
            # we need to append it to the previous entry's value
            if multiline_value:
                prev_node[prev_key] += f"\n{buffer}"
                buffer = ""

                # not sure if the next line continues a multiline value or if its the
                # start of a new top-level object. We assume it's not until we hit
                # an invalid character for a new top-level key :(
                multiline_value = False
                continue

            # othwerwise what's in the buffer is a single-line string value that we want to save
            curr_node[curr_key] = system_state_value_to_type(buffer)

            # saving the current key/value in a temporary structure in
            # case it's actually a multiline one so we can update it later if needed
            prev_node = curr_node
            prev_key = curr_key

            # starting a new top-level entry
            buffer = ""
            curr_key = ""
            reading_top_key = True

            continue

        # An open bracket, brace, or parens means we're starting a new sub-object
        if c in ["{", "[", "("] and not reading_key and not multiline_value:

            # ..but only if it's followed by a space. If it's not, we're still reading characters
            # into our buffer
            next_c = stream.read(1)
            if next_c != " ":
                buffer += c + next_c
                continue

            if buffer:
                raise ValueError(
                    f"Unexpected '{c}' at {stream.tell()}. Buffer: '{buffer}'"
                )

            obj_stack.append((curr_node, curr_key))

            if c == "{":
                curr_node = {}

            if c == "[":
                curr_node = []
            if c == "(":
                curr_node = MutableTuple()

            curr_key = ""
            continue

        # a single quote indicates the start or end of a sub-object dictionary key
        if c == "'" and not reading_key:
            if not isinstance(curr_node, dict):
                raise ValueError(f"unexpected '{c}' at {stream.tell()}")
            reading_key = True
            continue

        # if we're done reading in a sub-key we expect there to be a colon and a space
        if c == "'" and reading_key:

            if stream.read(2) != ": ":
                raise ValueError(
                    f"Expect colon after dict key {buffer} at {stream.tell()}"
                )
            curr_key = buffer
            buffer = ""
            reading_key = False
            continue

        # A comma often means we've finished reading a value in a sub-object (dict/list/tuple)...
        if c == ",":

            # ..but only if it's followed by a space. If it's not, we're still reading characters
            # into our buffer
            next_c = stream.read(1)
            if next_c != " ":
                buffer += c + next_c
                continue

            # if the next character was a space we need to save it to our current object
            if isinstance(curr_node, dict) and reading_key:
                raise ValueError(f"Unexpected comma at {stream.tell()}")

            if isinstance(curr_node, list) and buffer:
                curr_node.append(system_state_value_to_type(buffer))
            elif isinstance(curr_node, dict) and curr_key:
                curr_node[curr_key] = system_state_value_to_type(buffer)

            buffer = ""
            continue

        # A close bracket, brace, or parens means we're ending an object...
        if c in ["}", "]", ")"] and not (multiline_value or reading_key):

            # but only if the next two characters are a comma followed by a space,
            # or if the next character is a newline. If not, this is just
            # part of our value and we want to save it to the buffer and move on.
            next_c = stream.read(2)
            if len(next_c) == 2 and next_c != ", " and next_c[0] != "\n":
                buffer += c + next_c
                continue

            if c == "}" and not isinstance(curr_node, dict):
                raise ValueError(f"Unexpected '{c}' at {stream.tell()}")
            if c == "]" and not isinstance(curr_node, list):
                raise ValueError(f"Unexpected '{c}' at {stream.tell()}")
            if c == ")" and not isinstance(curr_node, MutableTuple):
                raise ValueError(f"Unexpected '{c}' at {stream.tell()}")

            # convert our 'mutable' tuple to a real one
            if isinstance(curr_node, MutableTuple):
                curr_node = tuple(curr_node)

            # save this object to the parent (either append or update as applicable)
            (parent_node, parent_key) = obj_stack.pop()
            if parent_key:
                parent_node[parent_key] = curr_node
            else:
                parent_node.append(curr_node)

            # set parent to current node
            curr_node = parent_node
            curr_key = ""

            # if the character after the close bracket/brace etc was a newline
            # and we were at the top of the stack, we started reading in
            # a new top-level object's key (one character past the newline)
            if len(next_c) == 2 and next_c[0] == "\n" and len(obj_stack) == 0:
                buffer += next_c[1]
                reading_top_key = True

            continue

        # add this key to our buffer

        buffer += c

    return curr_node



from typing import Union
from datetime import timedelta
import re
import logging
import sys
from ipaddress import ip_network, ip_interface, ip_address
from ipaddress import (
    IPv4Address,
    IPv6Address,
    IPv4Interface,
    IPv6Interface,
    IPv4Network,
    IPv6Network,
)
import json
from typing import List
from copy import deepcopy
from .models import RouteDict


# eventually want to pull this from a more central place
# as we've defined it in so many places already
INTERFACE_ABBRS = {
    # ios
    "Fastethernet": "Fa",
    "GigabitEthernet": "Gi",
    "TwoGigabitEthernet": "Tw",
    "TenGigabitEthernet": "Te",
    "TwentyFiveGigE": "Twe",
    "Port-channel": "Po",
    "Loopback": "Lo",
    # nxos
    "Ethernet": "Eth",
    "port-channel": "Po",
    "loopback": "Lo",
    "Management": "Mgmt",
}


class UMnetNapalmJsonEncoder(json.JSONEncoder):
    """
    custom json encoder that handles types
    in our results that aren't encodable by default.
    Reference the "Extending JSONEncoder" section of: https://docs.python.org/3/library/json.html

    Currently ipaddress and timedelta types are the only non-encodable
    types among the results
    """

    def default(self, obj):
        if isinstance(
            obj,
            (
                IPv4Address,
                IPv6Address,
                IPv4Interface,
                IPv6Interface,
                IPv4Network,
                IPv6Network,
                timedelta,
            ),
        ):
            return str(obj)

        return super().default(obj)


def abbr_interface(interface: str) -> str:
    """
    Converts long version of interface name to short one
    if applicable
    """
    for long, short in INTERFACE_ABBRS.items():
        if interface.startswith(long):
            return interface.replace(long, short)

    return interface


def age_to_integer(age: str) -> Union[int, None]:
    """
    Across platforms age strings can be:
    10y5w, 5w4d, 05d04h, 01:10:12, 3w4d 01:02:03,
    50 days 10 hours (ASA)
    """

    # convert empty string to none
    if age == "":
        return None

    # integer age is seconds - we don't need further parsing
    # if we got an integer value
    m = re.search(r"^\d+$", age)
    if m:
        return int(age)

    # Extracting digits out of different time strings
    TIME_MATCHES = [
        r"(?P<hours>\d+):(?P<minutes>\d+):(?P<seconds>\d+)",
        r"(?P<years>\d+) years?",
        r"(?P<months>\d+) months?",
        r"(?P<weeks>\d+) weeks?",
        r"(?P<days>\d+) days?",
        r"(?P<hours>\d+) hours?",
        r"(?P<minutes>\d+) minutes?",
        r"(?P<years>\d+)y",
        r"(?P<weeks>\d+)w",
        r"(?P<days>\d+)d",
        r"(?P<hours>\d+)h",
    ]
    # Multipliers for different units
    UNIT_TO_SECONDS = {
        "years": 31536000,
        "months": 2628000,
        "weeks": 604800,
        "days": 86400,
        "hours": 3600,
        "minutes": 60,
        "seconds": 1,
    }

    seconds = 0

    # search for every time match regex. If  a match is found
    # extract data, apply multiplier, and add to our seconds total
    for match in TIME_MATCHES:
        m = re.search(match, age)
        if not m:
            continue

        for unit, multiplier in UNIT_TO_SECONDS.items():
            if unit in m.groupdict():
                seconds += int(m.groupdict()[unit]) * multiplier

    return seconds


def resolve_nh(nh_ip: str, nh_table: str, routes: List[RouteDict]) -> List[RouteDict]:
    """
    Given a next hop IP and next hop table, search a list of
    routes for that next hop and return LPMs that have a next hop
    interface
    """
    nh_routes = [
        r
        for r in routes
        if r["vrf"] == nh_table
        and r["prefix"].version == ip_network(nh_ip).version
        and r["prefix"].supernet_of(ip_network(nh_ip))
        and r["nh_interface"]
    ]

    if not nh_routes:
        return []

    nh_routes.sort(key=lambda x: x["prefix"].prefixlen, reverse=True)

    # the first entry is a longest prefix match because of our sort,
    # but with ECMP there could be more than one!
    # we'll peel routes off the front with matching lengths to find
    # the rest.
    lpms = [nh_routes[0]]
    if len(nh_routes) > 1:
        for nh_route in nh_routes[1::]:
            if nh_route["prefix"].prefixlen == nh_routes[0]["prefix"].prefixlen:
                lpms.append(nh_route)

    return lpms


def resolve_all_nhs(routes: List[RouteDict], raise_on_error=False) -> List[RouteDict]:
    """
    Goes through a list of routes and resolves the interface for any
    next hops that don't already have them listed.
    """
    output = []
    for route in routes:
        if route["nh_interface"]:
            output.append(route)
            continue

        # attempt to resolve next hop
        nh_table = (
            "default" if route["mpls_label"] or route["vxlan_vni"] else route["vrf"]
        )
        nh_lpms = resolve_nh(route["nh_ip"], nh_table, routes)
        if not nh_lpms and raise_on_error:
            raise LookupError(f"Could not resolve next hop for {route}")

        # for every lpm found create an entry in our table for this recursive
        # route
        for lpm in nh_lpms:
            resolved = deepcopy(route)
            resolved["nh_ip"] = lpm["nh_ip"]
            resolved["nh_interface"] = lpm["nh_interface"]
            resolved["mpls_label"].extend(lpm["mpls_label"])
            output.append(resolved)

    return output


def str_to_type(string: str) -> object:
    """
    Converts string output to obvious types.
    empty quote is converted to "None".
    ip_address, ip_interface, and ip_network are converted to their appropriate objecst
    integers are converted to int
    """
    if string in ["", "None"]:
        return None
    if re.match(r"^\d+$", string):
        return int(string)

    for ip_type in [ip_network, ip_interface, ip_address]:
        try:
            return ip_type(string)
        except ValueError:
            pass

    return string


LOG_LEVELS = {
    "DEBUG": logging.DEBUG,
    "INFO": logging.INFO,
    "WARNING": logging.WARNING,
    "ERROR": logging.WARNING,
    "CRITICAL": logging.CRITICAL,
}
LOG_FORMAT = "%(asctime)-15s  %(levelname)8s %(name)s %(message)s"


def configure_logging(
    log_level: Union[int, str],
    log_globally: bool = False,
    log_file: str = None,
    log_to_console: bool = False,
):
    """
    Configures logging for the module, or globally as indicated by the input
    """

    if log_globally:
        logger = logging.getLogger()
    else:
        module_name = __name__.split(".")[0]
        logger = logging.getLogger(module_name)

    if isinstance(log_level, str):
        log_level = LOG_LEVELS[log_level.upper()]
    logger.setLevel(log_level)

    if log_file:
        file_handler = logging.handlers.RotatingFileHandler(
            log_file, maxBytes=1024 * 1024 * 10, backupCount=20
        )
        file_handler.setFormatter(logging.Formatter(LOG_FORMAT))
        logger.addHandler(file_handler)

    if log_to_console:
        stdout_handler = logging.StreamHandler(sys.stdout)
        stdout_handler.setFormatter(logging.Formatter(LOG_FORMAT))
        logger.addHandler(stdout_handler)



from typing import List, Dict
import re


from .models import (
    IPInterfaceDict,
    RouteDict,
    MPLSDict,
    VNIDict,
    InventoryDict,
    LagInterfaceDict,
    VALID_INVENTORY_TYPES,
)


class UMnetNapalmError(Exception):
    """
    Generic error class
    """


class UMnetNapalm:
    """
    Base class for um-specific (but also non-vendor specific)
    implementations.
    """

    # populate in child classes
    IGNORE_INTERFACES = []
    LABEL_VALUE_MAP = {}
    INVENTORY_TO_TYPE = {}
    PROTOCOL_ABBRS = {}
    I_ABBRS = {}

    def _parse_protocol_abbr(self, abbr) -> str:
        if abbr in self.PROTOCOL_ABBRS:
            return self.PROTOCOL_ABBRS[abbr]

        raise UMnetNapalmError(f"Unknown IOS protocol abbr {abbr}")

    def _expand_i(self, interface: str) -> str:
        """
        Converts short interface to long (I_ABBRS defined
        in child classes)
        """
        for short, long in self.I_ABBRS.items():
            if re.match(short, interface):
                return re.sub(short, long, interface)

        return interface

    def _abbr_i(self, interface: str) -> str:
        """
        Converts long interface to short (I_ABBRS defined
        in child classes)
        """
        for short, long in self.I_ABBRS.items():
            if re.match(long, interface):
                return re.sub(long, short, interface)

        return interface

    def _ignore_interface(self, interface: str) -> bool:
        """
        Checks to see if this interface matches any regexes in our
        IGNORE_INTERFACE list (implemented) in child class
        """
        for i in self.IGNORE_INTERFACES:
            if re.match(i, interface):
                return True

        return False

    def _get_inventory_type(self, name: str) -> str:
        """
        Maps the name of the inventory item to its type
        """
        for pattern, inv_type in self.INVENTORY_TO_TYPE.items():
            if inv_type and inv_type not in VALID_INVENTORY_TYPES:
                raise UMnetNapalmError(f"Invalid Inventory type {inv_type}")
            if re.search(pattern, name):
                return inv_type

        raise UMnetNapalmError(f"Unknown inventory item {name}")

    def get_ip_interfaces(self) -> List[IPInterfaceDict]:
        raise NotImplementedError

    def get_active_routes(self) -> List[RouteDict]:
        """get active routes"""
        raise NotImplementedError

    def get_mpls_switching(self) -> List[MPLSDict]:
        """get mpls switching (the mpls forwarding table)"""
        raise NotImplementedError

    def get_vni_information(self) -> List[VNIDict]:
        """get vni to vlan and VRF mapping"""
        raise NotImplementedError

    def get_inventory(self) -> List[InventoryDict]:
        """get vni to vlan and VRF mapping"""
        raise NotImplementedError

    def get_lag_interfaces(self) -> Dict[str, LagInterfaceDict]:
        """get lag parents, their status, and their children"""
        raise NotImplementedError

    def _ignore_interface(self, interface: str) -> bool:
        for ignore in self.IGNORE_INTERFACES:
            if re.search(ignore, interface):
                return True
        return False

    def _parse_label_value(self, label) -> list:
        """
        Parses mpls label value into normalized data
        """
        if label in self.LABEL_VALUE_MAP:
            return self.LABEL_VALUE_MAP[label]

        return [label]
